"""Use wraith to compare current version against published docs.
"""

import unittest
import os
import copy
import re
import yaml
import subprocess
import contextlib
from distutils.version import LooseVersion
import http.server
import socketserver
import threading

REFERENCE_URL = "https://www.cgat.org/downloads/public/CGATReport/documentation"
WRAITH_WORKDIR = os.path.abspath("wraith")
TEST_PORT=9100
TEST_HOST="localhost"

spider_config_template = """
browser: "phantomjs"
domains:
  test: http://{test_host}:{test_port}
spider_skips:
  - !ruby/regexp /static$/
  - !ruby/regexp /%23/
  - !ruby/regexp /.eps$/
  - !ruby/regexp /.svg$/
  - !ruby/regexp /.xlsx$/
  - !ruby/regexp /notebook/
  - !ruby/regexp /code/
directory: 'shots'
imports: "{wraith_data_config}"
phantomjs_options: '--ignore-ssl-errors=true --ssl-protocol=tlsv1'
"""

capture_config_template = """
browser: "phantomjs"
domains:
  test: http://{test_host}:{test_port}
  current: {reference_url}
spider_skips:
  - !ruby/regexp /static$/
  - !ruby/regexp /%23/
imports: "{wraith_data_config}"
screen_widths:
  - 1280
directory: 'shots'
fuzz: '20%'
threshold: 5
gallery:
  thumb_width:  200
  thumb_height: 200
mode: diffs_only
phantomjs_options: '--ignore-ssl-errors=true --ssl-protocol=tlsv1'
"""

@contextlib.contextmanager
def changedir(path):
    save_dir = os.path.abspath(os.getcwd())
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(save_dir)


def run_server():
    run("python -m http.server {} >& server.log".format(TEST_PORT))

        
@contextlib.contextmanager
def start_server(workdir):

    handler = http.server.SimpleHTTPRequestHandler

    with changedir(workdir):
        # thread = threading.Thread(target=run_server)
        # thread.start()
        print("yielding")
        yield
        print("back from yield")


def run(statement,
        return_stdout=False,
        return_popen=False,
        **kwargs):
    '''execute a command line statement.

    By default this method returns the code returned by the executed
    command. If *return_stdout* is True, the contents of stdout are
    returned as a file object. If *return_popen*, the Popen object is
    returned.

    ``kwargs`` are passed on to subprocess.call,
    subprocess.check_output or subprocess.Popen.

    Raises
    ------

    OSError
       If process failed or was terminated.

    '''

    # remove new lines
    statement = " ".join(re.sub("\t+", " ", statement).split("\n")).strip()
    print(statement)
    
    if "<(" in statement:
        shell = os.environ.get('SHELL', "/bin/bash")
        if "bash" not in shell:
            raise ValueError(
                "require bash for advanced shell syntax: <()")
        # Note: pipes.quote is deprecated. In Py3, use shlex.quote
        # (not present in Py2.7)
        statement = "%s -c %s" % (shell, pipes.quote(statement))

    if return_stdout:
        return subprocess.check_output(statement, shell=True, **kwargs).decode("utf-8")
    elif return_popen:
        return subprocess.Popen(statement, shell=True, **kwargs)
    else:
        retcode = subprocess.call(statement, shell=True, **kwargs)
        if retcode < 0:
            raise OSError("process was terminated by signal %i" % -retcode)
        return retcode


def check_version(cmd, regex, min_version):
    
    version_txt = run(cmd , return_stdout=True)
    version = re.search(regex, version_txt).groups()[0]
    if LooseVersion(version) < LooseVersion(min_version):
        raise ValueError("version check failed: {} < {}, '{}'".format(
            version, min_version, cmd))
    
    return version


class TestWraith(unittest.TestCase):

    def setUp(self):

        source_dir = os.path.join(
            os.path.dirname(os.path.dirname(
                os.path.abspath(__file__))),
            "doc", "_build", "html")
        
        # check if npm is intalled
        npm_version = check_version("npm --version", "(\S+)", "3.10")

        # check if phantomjs is installed
        phantomjs_version = check_version("npm list -g | grep phantom",
                                          "phantomjs@(\S+)",
                                          "2.1")

        
        ruby_version = check_version("ruby --version",
                                     "ruby (\S+)",
                                     "2.1")
        
        wraith_version = check_version(
            "gem list | grep wraith",
            "wraith \((\S+)\)",
            "4.0.1")

        # get gem info
        gem_data = yaml.load(run("gem environment", return_stdout=True))
        gem_paths = []
        for record in gem_data["RubyGems Environment"]:
            for key, value in record.items():
                if key == "GEM PATHS":
                    gem_paths.extend(value)
                    break
        if not gem_paths:
            raise ValueError("could not find GEM PATHS in gem environment")

        filenames = [os.path.join(path,
                                  "gems/wraith-{}/lib/wraith/spider.rb".format(wraith_version))
                     for path in gem_paths]
        if sum([os.path.exists(fn) for fn in filenames]) == 0:
            raise ValueError("could not find file spider.rb to patch in {}".format(filenames))            
        
        for fn in filenames:

            if not os.path.exists(fn):
                continue
            
            with open(fn) as inf:
                data = inf.read()
                
            if "path.downcase" in data:
                with open(fn, "w") as outf:
                    outf.write(re.sub("path.downcase", "path", data))
                    
        # crawl new docs to collect documents to test
        config_dir = os.path.abspath(os.path.join(WRAITH_WORKDIR, "config"))
        wraith_spider_config = os.path.join(config_dir, "wraith_spider.yml")
        wraith_capture_config = os.path.join(config_dir, "wraith_capture.yml")
        wraith_data_config = os.path.join(config_dir, "wraith_data.yml")
        
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)

        if not os.path.exists(wraith_spider_config):
            # do not crawl with reference, as crawler follows external links
            spider_config = spider_config_template.format(
                wraith_data_config=os.path.basename(wraith_data_config),
                test_host=TEST_HOST,
                test_port=TEST_PORT)
            with open(wraith_spider_config, "w") as outf:
                outf.write(spider_config)

        if not os.path.exists(wraith_data_config):
            with start_server(source_dir) as server:
                run("cd {} && wraith spider {}".format(WRAITH_WORKDIR, wraith_spider_config))

        if not os.path.exists(wraith_capture_config):
            # do not crawl with reference, as crawler follows external links
            capture_config = capture_config_template.format(
                wraith_data_config=os.path.basename(wraith_data_config),
                reference_url=REFERENCE_URL,
                test_host=TEST_HOST,
                test_port=TEST_PORT)
            with open(wraith_capture_config, "w") as outf:
                outf.write(capture_config)

        self.wraith_capture_config = wraith_capture_config
        self.source_dir = source_dir
        
    def test_against_reference(self):
        with start_server(self.source_dir) as server:
            run("cd {} && wraith capture {}".format(WRAITH_WORKDIR,
                                                    self.wraith_capture_config))


if __name__ == "__main__":
    unittest.main()

        



import os
import warnings

import matplotlib
import matplotlib.pyplot as plt
from matplotlib import _pylab_helpers

from CGATReport import Utils
from CGATReport.Plugins.Collector import Collector


class MatplotlibPlugin(Collector):

    def __init__(self, *args, **kwargs):
        Collector.__init__(self, *args, **kwargs)

        plt.close('all')
        # matplotlib.rcdefaults()
        # set a figure size that doesn't overflow typical browser windows
        matplotlib.rcParams['figure.figsize'] = (5.5, 4.5)

    def collect(self, blocks, figure_key=None, subfig=0):
        '''collect one or more matplotlib figures and

        1. save as png, hires-png and pdf
        2. save thumbnail
        3. insert rendering code at placeholders in output

        returns a map of place holder to placeholder text.
        '''
        fig_managers = _pylab_helpers.Gcf.get_all_fig_managers()

        self.debug("figure_key: {}, subfig: {}: collecting {} matplotlib images".format(
            figure_key, subfig, len(fig_managers)))

        map_figure2text = {}

        # determine the image formats to create
        default_format, additional_formats = Utils.getImageFormats(
            self.display_options)
        all_formats = [default_format, ] + additional_formats

        # create all required images
        for figman in fig_managers:

            # create all images
            figid = figman.num

            # select the correct figure
            figure = plt.figure(figid)

            # save explicit formats
            outname = "%s_%s_%02d_%02d" % (self.template_name, figure_key, subfig, figid)

            has_output = False

            for id, format, dpi in all_formats:

                outpath = os.path.join(self.outdir, '%s.%s' % (outname, format))

                # sanitize figure size for Agg.
                if figure.get_figwidth() > 32768 or figure.get_figheight() > 32768:
                    warnings.warn(
                        "figure size unexpected large {} x {}, "
                        "patching to 10x10 inches".format(
                            figure.get_figwidth(), figure.get_figheight()))
                    figure.set_size_inches(10, 10)

                try:
                    figman.canvas.figure.savefig(outpath, dpi=dpi)
                except Exception as ex:
                    s = Utils.collectExceptionAsString(
                        "exception raised while building plot '%s': %s" %
                        (outpath, str(ex)))
                    warnings.warn(s)
                    continue
                has_output = True

            if not has_output:
                warnings.warn("no output for '%s'" % outpath)
                continue

            # insert display figure
            is_html = False
            script_text = None

            # release memory
            plt.close(figure)

            # create the text element
            rst_output = Utils.buildRstWithImage(
                outname,
                self.outdir,
                self.rstdir,
                self.builddir,
                self.srcdir,
                additional_formats,
                self.tracker_id,
                self.links,
                self.display_options,
                default_format,
                is_html=is_html,
                text=script_text)

            if figure_key is None:
                f = figid
            else:
                f = "{}.{}".format(figure_key, figid)
            map_figure2text["#$mpl {}$#".format(f)] = rst_output

        return map_figure2text



import os
import random
import glob

from CGATReport.Tracker import Tracker
from collections import OrderedDict as odict
from CGATReport.ResultBlock import ResultBlock, ResultBlocks
from CGATReport import Utils

from matplotlib import pyplot as plt

try:
    from rpy2.robjects import r as R
except ImportError:
    R = None


class MatplotlibData(Tracker):

    '''create plot using matplotlib.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice):
        s = [random.randint(0, 20) for x in range(40)]
        random.shuffle(s)

        # do the plotting
        fig = plt.figure()
        plt.plot(s)
        return {"text": "#$mpl %i$#" % fig.number}


class RPlotData(Tracker):

    '''create plot using R.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        if not R:
            return

        s = [random.randint(0, 20) for x in range(40)]
        random.shuffle(s)
        # do the plotting
        R.x11()
        R.plot(s, s)

        return {"text": "#$rpl %i$#" % R["dev.cur"]()[0]}


IMAGEDIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "..", "images")


class Images(Tracker):

    tracks = glob.glob(os.path.join(IMAGEDIR, "*.png"))

    def __call__(self, track, slice=None):

        rst_text = '''
This is a preface

.. figure:: %s

Some more text for the figure\n''' % track

        return {"rst": rst_text}


class Images2(Tracker):

    tracks = "all"

    def __call__(self, track, slice=None):

        blocks = ResultBlocks()

        block = '''
.. figure:: %(image)s
   :height: 300
'''
        for image in glob.glob(os.path.join(IMAGEDIR, "*.png")):
            blocks.append(ResultBlock(text=block % locals(),
                                      title="image"))

        return odict((("rst", "\n".join(
            Utils.layoutBlocks(blocks, layout="columns-2"))),))



import os
from CGATReport.Plugins.Collector import Collector
from CGATReport import Utils

try:
    import holoviews as hv
    from holoviews.plotting import mpl
    HAS_HOLOVIEW = True
except ImportError:
    HAS_HOLOVIEW = False


try:
    import bokeh.embed
    import bokeh.plotting
    HAS_BOKEH = True
except ImportError:
    HAS_BOKEH = False
    

class HoloviewsPlugin(Collector):

    engine = "bokeh"
    
    def __init__(self, *args, **kwargs):
        Collector.__init__(self, *args, **kwargs)

    def collect(self, blocks, figure_key="", subfig=0):
        '''collect holoview output.

        '''
        map_figure2text = {}
        extension = "svg"
        rst_text = '''.. figure:: /{absfn}
'''

        for block in blocks:
            if not hasattr(block, "hv"):
                continue

            if self.engine == "mpl":
                renderer = hv.Store.renderers['matplotlib'].instance(fig='svg')
                plot = renderer.get_plot(block.hv)
                rendered = renderer(plot, fmt="auto")
        
                (data, info) = rendered
            
                # remove special characters from filename. I think the docutils
                # system removes these from links which later causes problems
                # as the link does not point to the correct location of the
                # file.
                outname = Utils.quote_filename(
                    "%s_%s" % (self.template_name, block.title))
                outputpath = os.path.join(
                    self.outdir, '%s.%s' % (outname, extension))

                # save to file
                with open(outputpath, "w") as outf:
                    outf.write(data)

                # use absolute path
                absfn = os.path.abspath(outputpath)
                map_figure2text["#$hv %s$#" % block.title] = rst_text.format(
                    absfn=absfn)
            elif self.engine == "bokeh":
                renderer = hv.renderer("bokeh")
                html = renderer.static_html(block.hv)

                lines = [".. only:: html\n"] +\
                        ["   .. raw:: html\n"] +\
                        ["      " + x for x in html.splitlines()]

                lines = "\n".join(lines)
                
                map_figure2text["#$hv %s$#" % block.title] = lines

        return map_figure2text



from CGATReport.Component import Component


class Collector(Component):

    capabilities = ['collect']

    def __init__(self, *args, **kwargs):
        Component.__init__(self, *args, **kwargs)

        self.template_name = kwargs.get("template_name")
        self.outdir = kwargs.get("outdir")
        self.rstdir = kwargs.get("rstdir")
        self.builddir = kwargs.get("builddir")
        self.srcdir = kwargs.get("srcdir")
        self.content = kwargs.get("content")
        self.display_options = kwargs.get("display_options")
        self.tracker_id = kwargs.get("tracker_id")
        self.links = kwargs.get("links")



import os
import re
from CGATReport.Plugins.Collector import Collector


class RSTPlugin(Collector):
    '''collect rst text.

    This plugin looks for image/figure directives in literal text
    output and changes absolute paths to relative paths so that they
    are valid within sphinx.

    images: path relative to rst directory

    links: path relative to html content

    '''

    capabilities = ['collect']

    # include white spaces at the end for length-neutral substitution
    # Pattern will also include a terminating '"'. This is used to
    # test if the context is a csv table, in which case no padding
    # should be done.
    # Ignore images with absolute path.
    rx_img = re.compile("\.\. (image|figure):: ([^/][^ |+,:]+[ ]*)")

    # external link targets:
    # .. _link: target
    rx_link = re.compile("\.\. _([^ |+,:]+)\s*:\s*([^ |+,:]+)[ ]*")

    def __init__(self, *args, **kwargs):
        Collector.__init__(self, *args, **kwargs)

    def collect(self, blocks, figure_key="", subfig=0):
        '''collect rst output from result blocks.

        '''
        map_figure2text = {}

        def replace_and_pad(s, old, new):
            '''replace old with new in s such that length of
            new+spaces is at least that of old+spaces.

            There needs to be enough space for padding. If there is not
            enough space, the new string will be truncated and an error
            issued.
            '''
            # do not pad in csv tables
            if not old.endswith('"'):
                oldlen = len(old)
                newlen = len(new)
                new = new + " " * (oldlen - newlen)

            if len(new) > len(old):
                self.warn("length of substitution string ({}) is "
                          "longer than original ({}):\n{}\n{}".format(
                              len(new), len(old),
                              new, old))
                new = new[:len(old)]

            n = s.replace(old, new)
            return n

        for block in blocks:
            if not hasattr(block, "text"):
                continue
            lines = block.text.split("\n")
            n = []
            for l in lines:
                ll = l
                # for x in self.rx_img.finditer(ll):
                #     directive, filename = x.groups()
                #     relpath = os.path.relpath(
                #         filename.strip(),
                #         os.path.abspath(self.rstdir))
                #     newpath = re.sub("\\\\", "/", relpath)
                #     l = replace_and_pad(l, filename, newpath)

                for x in self.rx_link.finditer(ll):
                    directive, filename = x.groups()
                    newpath = re.sub(
                        "\\\\", "/",
                        os.path.abspath(filename.strip()))
                    # pad with spaces to keep table alignment
                    l = replace_and_pad(l, filename, newpath)

                n.append(l)
            block.text = "\n".join(n)

        return map_figure2text



import os
from CGATReport.Component import Component
from CGATReport.Utils import indent

try:
    import bokeh.embed
    import bokeh.plotting
    HAS_BOKEH = True
except ImportError:
    HAS_BOKEH = False


class BokehPlugin(Component):
    '''collect bokeh plot objects and write
    javascript snippets to be included
    raw into html.
    '''

    capabilities = ['collect']

    def __init__(self, *args, **kwargs):
        Component.__init__(self, *args, **kwargs)

    def collect(self,
                blocks, figure_key="", subfig=0):
        '''collect rst output from result blocks.

        '''

        map_figure2text = {}

        if not HAS_BOKEH:
            return map_figure2text

        figid = 10

        for block in blocks:
            if not hasattr(block, "bokeh"):
                continue

            figid = block.bokeh._id
            res = bokeh.resources.CDN
            script_path = os.path.join(
                '_static/report_directive/', "%s.js" % figid)

            js_txt, script_txt = bokeh.embed.autoload_static(
                block.bokeh, res, script_path)

            # make sure to add the '/' at the end
            # script = block.bokeh.create_html_snippet(
            #     server=False,
            #     embed_base_url='_static/report_directive/',
            #     embed_save_loc=outdir,
            #     static_path='_static/')

            with open(script_path, "w") as outf:
                outf.write(js_txt)

            text = """.. raw:: html

   <div><p>
{}
   </p></div>
""".format(indent(script_txt, 8))

            map_figure2text["#$bkh %s$#" % figid] = text
        return map_figure2text



import os
from CGATReport.Plugins.Collector import Collector
from CGATReport import Utils


class XLSPlugin(Collector):

    capabilities = ['collect']

    def __init__(self, *args, **kwargs):
        Collector.__init__(self, *args, **kwargs)

    def collect(self, blocks, figure_key="", subfig=0):
        '''collect xls output from result blocks.

        xls output is written to a file and a link will be inserted at
        the place holder.
        '''
        map_figure2text = {}
        extension = "xlsx"
        for block in blocks:
            if not hasattr(block, "xls"):
                continue

            outname = Utils.quote_filename(
                "%s_%s" % (self.template_name, block.title))
            outputpath = os.path.join(self.outdir, '%s.%s' %
                                      (outname, extension))
            
            # save to file
            block.xls.save(outputpath)

            # use absolute path
            link = os.path.abspath(outputpath)

            rst_output = ":download:`(link) </{}>`".format(link)
            map_figure2text["#$xls %s$#" % block.title] = rst_output

        return map_figure2text



'''Plotting using the ggplot module.'''
import sys

from CGATReport.Plugins.Renderer import Renderer
from CGATReport.Plugins.Plotter import Plotter
from docutils.parsers.rst import directives

import matplotlib.pyplot as plt
import pandas
import copy
import warnings

try:
    from plotnine import *
except ImportError:
    pass


class GGPlot(Renderer, Plotter):
    """Use the python plotnine libary for plotting.
    """
    options = (
        ('aes',  directives.unchanged),
        ('geom', directives.unchanged),
    ) + Renderer.options + Plotter.options

    group_level = 0

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)
        Plotter.__init__(self, *args, **kwargs)

        if "aes" not in kwargs:
            raise ValueError("ggplot renderer requires an aesthetics option")

        if "geom" not in kwargs:
            raise ValueError("ggplot renderer requires a geom option")

        self.aes = kwargs.get('aes')
        self.geom = kwargs.get('geom')

    def render(self, dataframe, path):

        # the index in the dataframe is reset in order to add the
        # index as a column so that it is available for plotting.
        dataframe.reset_index(inplace=True)
        
        if len(dataframe.dropna()) == 0:
            warnings.warn("dataframe is empty after dropping all NA columns")
            return None

        s = "plot = ggplot(aes(%s), data=dataframe) + %s" % (self.aes, self.geom)
        ll = copy.copy(locals())
        gl = copy.copy(globals())
        try:
            exec(s, gl, ll)
        except Exception as msg:
            raise Exception(
                "plotnine raised error for statement '{}': msg={}".format(s, msg))
        plot = ll["plot"]

        if self.title:
            plt.title(self.title)

        try:
            plts = [plot.draw()]
        except Exception as msg:
            raise Exception(
                "ggplot raised error on rendering for statement '{}': msg={}".format(
                s, msg))
        return self.endPlot(plts, None, path)



import os
from CGATReport.Plugins.Collector import Collector
from CGATReport import Utils


class HTMLPlugin(Collector):

    def __init__(self, *args, **kwargs):
        Collector.__init__(self, *args, **kwargs)

    def collect(self, blocks, figure_key="", subfig=0):
        '''collect html output from result blocks.

        HTML output is written to a file and a link will be inserted at
        the place holder.
        '''
        map_figure2text = {}
        extension = "html"
        for block in blocks:
            if not hasattr(block, "html"):
                continue

            # remove special characters from filename. I think the docutils
            # system removes these from links which later causes problems
            # as the link does not point to the correct location of the
            # file.
            outname = Utils.quote_filename(
                "%s_%s" % (self.template_name, block.title))
            outputpath = os.path.join(
                self.outdir, '%s.%s' % (outname, extension))

            # save to file
            with open(outputpath, "w", encoding="utf-8") as outf:
                outf.write(block.html)

            # use absolute path
            link = os.path.abspath(outputpath)
            rst_output = ":download:`(link) </{}>`".format(link)
            map_figure2text["#$html %s$#" % block.title] = rst_output

        return map_figure2text



import re

from CGATReport.ResultBlock import ResultBlock, ResultBlocks
from CGATReport.Plugins.Renderer import Renderer
from CGATReport.DataTree import path2str
from docutils.parsers.rst import directives

try:
    import holoviews as hv
    from holoviews.plotting import mpl
    HAS_HOLOVIEW = True
except ImportError:
    HAS_HOLOVIEW = False


class HoloviewsPlot(Renderer):

    options = Renderer.options + \
              (('statement', directives.unchanged),)

    statement = None
    engine = "bokeh"
    
    def __init__(self, *args, **kwargs):
        self.statement = kwargs.get("statement", "")
        if self.statement is None:
            raise ValueError("HoloviewPlot requires 'statement' to be given")

        if not HAS_HOLOVIEW:
            raise ValueError("Holoview can not be imported")
        
    def render(self, data, path):
        """execute holoview statements."""
        statement = self.statement.strip()
        assert statement.endswith(")")

        g, l = globals(), locals()
        layout = None
        try:
            exec("layout = %s" % statement, g, l)
        except Exception as msg:
            raise Exception(
                "holoviews raised error for statement '%s': msg=%s" %
                (statement, msg))

        return self.endPlot(l["layout"], None, path)

    def endPlot(self, layout, legend, path):
        title = path2str(path)
        r = ResultBlock(
            text="#$hv {}$#".format(title),
            title=title)
        r.hv = layout
        
        return ResultBlocks(r)
        
